// Copyright (c) 2022 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#pragma once
#include "dawn.hpp"
#include "asset/AssetLoader.hpp"

namespace Dawn {
  class AssetManager final : public std::enable_shared_from_this<AssetManager> {
    private:
      std::vector<std::shared_ptr<AssetLoader>> pendingAssetLoaders;
      std::vector<std::shared_ptr<AssetLoader>> finishedAssetLoaders;

      /**
       * Returns an existing asset loader if it exists.
       * 
       * @param filename The filename of the asset to get.
       * @return The asset loader if it exists, otherwise nullptr.
       */
      template<class T>
      std::shared_ptr<T> getExisting(const std::string &filename) {
        auto existing = std::find_if(
          pendingAssetLoaders.begin(),
          pendingAssetLoaders.end(),
          [&](auto &loader) {
            return (
              loader->name == filename &&
              loader->getAssetType() == T::ASSET_TYPE
            );
          }
        );

        if(existing == pendingAssetLoaders.end()) {
          existing = std::find_if(
            finishedAssetLoaders.begin(), finishedAssetLoaders.end(),
            [&](auto &loader) {
              return (
                loader->name == filename &&
                loader->getAssetType() == T::ASSET_TYPE
              );
            }
          );
          
          if(existing == finishedAssetLoaders.end()) return nullptr;
        }

        return std::static_pointer_cast<T>(*existing);
      }

      /**
       * Removes an existing asset loader if it exists.
       * 
       * @param filename The filename of the asset to remove.
       */
      void removeExisting(const std::string &filename);

    public:
      /**
       * Initializes this asset manager so it can begin accepting assets.
       */
      void init();

      /**
       * Updates the asset manager.
       */
      void update();

      /**
       * Returns whether the asset manager has loaded all of the currently
       * managed assets.
       * 
       * @return True if all assets have been loaded.
       */
      bool_t isEverythingLoaded();

      /**
       * Returns the asset loader for the given asset.
       * 
       * @param filename The filename of the asset to get.
       * @return The asset loader for the given asset.
       */
      template<class T>
      std::shared_ptr<T> get(const std::string &filename) {
        auto existing = this->getExisting<T>(filename);
        if(existing) return existing;

        std::shared_ptr<T> loader = std::make_shared<T>(
          shared_from_this(),
          filename
        );
        pendingAssetLoaders.push_back(
          std::static_pointer_cast<AssetLoader>(loader)
        );
        return loader;
      }

      /**
       * Dispose the asset manager, and all attached assets.
       */
      ~AssetManager();
  };
}