// Copyright (c) 2023 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#include "scene/SceneItem.hpp"
#include "scene/Scene.hpp"

using namespace Dawn;

SceneItem::SceneItem(const std::weak_ptr<Scene> scene) :
  scene(scene),
  SceneItemTransform(),
  SceneItemComponents()
{
}

std::shared_ptr<SceneItem> SceneItem::sceneItemComponentsSelf() {
  return shared_from_this();
}

std::shared_ptr<Scene> SceneItem::getScene() {
  auto s = scene.lock();
  assertNotNull(s, "Scene has unloaded?");
  return s;
}

void SceneItem::init() {
  auto sharedThis = shared_from_this();

  // Loop until all components initialized...
  while(true) {
    // Create copy of the components, components may chose to add more components
    // but those sub components will not be initialized at this time.
    auto components = this->components;
    for(auto &component : components) {
      if(component->isInitialized()) continue;
      component->init(sharedThis);
    }

    // If they are all initalized we are fine.
    components = this->components;
    bool_t allInitialized = std::all_of(
      components.begin(),
      components.end(),
      [](auto &component) {
        return component->isInitialized();
      }
    );

    if(allInitialized) break;
  }
}

void SceneItem::deinit() {
  // Create copy of the components, components may chose to add more components
  // but those sub components will not be disposed at this time.
  auto components = this->components;

  for(auto &component : components) {
    if(!component->isInitialized()) continue;
    component->dispose();
  }

  this->components.clear();
}

void SceneItem::remove() {
  auto scene = getScene();
  if(!scene) return;
  scene->removeItem(shared_from_this());
}

SceneItem::~SceneItem() {
  this->deinit();
}